import 'dart:async';
import 'package:bloc/bloc.dart';
import 'package:relations/utils/node_arrangement.dart';
import '../models/graph_edge.dart';
import '../models/graph_node.dart';
import 'package:flutter/widgets.dart';
import './bloc.dart';
import '../models/graph.dart';
import 'package:rxdart/rxdart.dart';

enum EdgeEventType {
  none,
  nodeMoved,
  nodeCreated
}

class EdgeEvent {
  EdgeEventType type;
  String edgeID;
  EdgeEvent(this.type, this.edgeID);
}

class GraphBloc extends Bloc<GraphEvent, GraphState> {

  Graph graph;

  PublishSubject<EdgeEvent> edgeNotifier;
  Observable<EdgeEvent> get edgeNotifierStream => edgeNotifier.stream;

  GraphBloc({@required this.graph}) : super() {
    edgeNotifier = PublishSubject<EdgeEvent>();
  }

  @override
  GraphState get initialState => GraphLoadedState(graph.nodes, graph.allEdges);

  @override
  Stream<GraphState> mapEventToState(GraphEvent event) async* {
    if (event is AddNodeEvent) {
      GraphState state = currentState;
      if (state is GraphLoadedState) {
        graph.addNode(event.node);
        final updatedNodes = List<GraphNode>.from(graph.nodes);
        yield GraphLoadedState(updatedNodes, graph.allEdges);
      }
    } 
    else
    if (event is RemoveNodeEvent) {
      graph.removeNodeByID(event.nodeID);
      yield GraphLoadedState(graph.nodes, graph.allEdges);
    }
    else 
    if (event is ConnectNodesEvent) {
      GraphState state = currentState;
      if (state is GraphLoadedState) {
         if (graph.connectNodes(event.firstNodeID, event.secondNodeID)) {
            yield GraphLoadedState(state.nodes, graph.allEdges);
         }
      }
    }
    else
    if (event is DisconnectNodesEvent) {
      GraphState state = currentState;
      if (state is GraphLoadedState) {
        graph.disconnectNodes(event.firstNodeID, event.secondNodeID);
        yield GraphLoadedState(state.nodes, graph.allEdges);
      }
    }
    else
    if (event is NodeArrangementEvent) {

      GraphState state = currentState;

      if (state is GraphLoadedState) {

        List<GraphNode> arrangedNodes;

        if (event.arrangementType == NodeArrange.circle) {
          arrangedNodes = List<GraphNode>.from(NodeArrangement.circle(event.center, graph.nodes));
        }
        else
        if (event.arrangementType == NodeArrange.matrix) {
          arrangedNodes = List<GraphNode>.from(NodeArrangement.matrix(event.center, graph.nodes));
        }
        else
        if (event.arrangementType == NodeArrange.horizontal) {
          arrangedNodes = List<GraphNode>.from(NodeArrangement.horizontal(event.center, graph.nodes));
        }
        else
        if (event.arrangementType == NodeArrange.vertical) {
          arrangedNodes = List<GraphNode>.from(NodeArrangement.vertical(event.center, graph.nodes));
        }
        
        yield GraphLoadedState(arrangedNodes, graph.allEdges);
      }
    }
  }

  void notifyConnectedEdges(GraphNode node) {
    List<GraphEdge> connected = graph.getEdges(node.id);
    
    connected?.forEach((GraphEdge e) {
      edgeNotifier.add(EdgeEvent(EdgeEventType.nodeMoved, e.id));
    });
  }

  @override
  void dispose() {
    super.dispose();
    this.edgeNotifier.close();
  }
}
